import flux
import functools
import logbook
import sys
import waiting
from ._compat import reraise
from waiting.exceptions import TimeoutExpired

_logger = logbook.Logger(__name__)


class PactBase(object):

    def __init__(self, timeout_seconds):
        super(PactBase, self).__init__()
        self._triggered = False # Make sure we only trigger 'then' once, even if using greenlets
        self._finished = False
        self._then = []
        self._lastly = []
        self._during = []
        self._timeout_callbacks = []
        self._end_time = None if timeout_seconds is None else flux.current_timeline.time() + timeout_seconds
        _logger.debug("{0!r} was created", self)

    def _validate_can_add_callback(self):
        if self._finished:
            raise RuntimeError('Cannot append then callbacks after pact was finished')

    def get_timeout_exception(self, exc_info): # pylint: disable=no-self-use, unused-argument
        """Returns exception to be used when wait() times out. Default reraises waiting.TimeoutExpired.
        """
        return None

    def is_finished(self):
        """Returns whether or not this pact is finished
        """
        return self._finished

    def poll(self):
        """Checks all predicates to see if we're finished
        """
        if self._finished:
            return True

        exc_info = self._process_callbacks('during')

        self._finished = self._is_finished()

        if self._finished and not self._triggered:
            self._triggered = True
            then_exc_info = self._process_callbacks('then')
            lastly_exc_info = self._process_callbacks('lastly')
            exc_info = exc_info or then_exc_info or lastly_exc_info

        if exc_info is not None:
            reraise(*exc_info)

        return self.is_finished()

    def _process_callbacks(self, name):
        exc_info = None
        for callback in getattr(self, '_{}'.format(name)):
            try:
                callback()
            except Exception: # pylint: disable=broad-except
                if exc_info is None:
                    exc_info = sys.exc_info()
                _logger.debug("Exception thrown from '{}' callback {!r} of {!r}", name, callback, self, exc_info=True)
        return exc_info

    def then(self, callback, *args, **kwargs):
        """Calls ``callback`` when this pact is finished
        """
        assert callable(callback)
        self._validate_can_add_callback()
        self._then.append(functools.partial(callback, *args, **kwargs))
        return self

    def lastly(self, callback, *args, **kwargs):
        """Calls ``callback`` when this pact is finished, after all 'then' callbacks
        """
        assert callable(callback)
        self._validate_can_add_callback()
        self._lastly.append(functools.partial(callback, *args, **kwargs))
        return self

    def during(self, callback, *args, **kwargs):
        """Calls ``callback`` periodically while waiting for the pact to finish
        """
        assert callable(callback)
        self._validate_can_add_callback()
        self._during.append(functools.partial(callback, *args, **kwargs))
        return self

    def on_timeout(self, callback, *args, **kwargs):
        """Calls ``callback`` when a wait timeout is encountered
        """
        assert callable(callback)
        self._validate_can_add_callback()
        self._timeout_callbacks.append(functools.partial(callback, *args, **kwargs))
        return self

    def _is_finished(self):
        raise NotImplementedError()  # pragma: no cover

    def wait(self, **kwargs):
        """Waits for this pact to finish
        """
        _logger.debug("Waiting for {!r}...", self)
        start_time = flux.current_timeline.time()
        if 'timeout_seconds' not in kwargs and self._end_time is not None:
            kwargs['timeout_seconds'] = max(0, self._end_time - start_time)
        try:
            waiting.wait(self.poll, waiting_for=self, **kwargs)
        except TimeoutExpired:
            exc_info = sys.exc_info()
            for timeout_callback in self._timeout_callbacks:
                timeout_callback()
            exc = self.get_timeout_exception(exc_info)  # pylint: disable=assignment-from-none
            if exc is None:
                reraise(*exc_info)
            else:
                raise exc # pylint: disable=raising-bad-type
        except Exception:
            _logger.debug("Exception was raised while waiting for {!r}", self, exc_info=True)
            raise
        else:
            finish_time = flux.current_timeline.time()
            _logger.debug("Finished waiting for {!r} (took: {:.04f} sec)", self, finish_time - start_time)

    def __repr__(self):
        raise NotImplementedError() # pragma: no cover



import flux
from forge import Forge

import pytest
from pact import Pact

# pylint: disable=redefined-outer-name

@pytest.fixture
def callback(forge):
    return forge.create_wildcard_function_stub()


class Pred(object):

    _satisfied = False

    def satisfy(self):
        self._satisfied = True

    def __call__(self):
        return self._satisfied


class Checkpoint(object):

    called_times = 0

    def __call__(self):
        self.called_times += 1

    @property
    def called(self):
        return self.called_times > 0


@pytest.fixture
def checkpoint():
    return Checkpoint()


@pytest.fixture
def checkpoint1():
    return Checkpoint()


@pytest.fixture
def checkpoint2():
    return Checkpoint()


@pytest.fixture
def checkpoint3():
    return Checkpoint()


@pytest.fixture
def timed_predicate_factory(deadline):
    def factory():
        return lambda: flux.current_timeline.time() >= deadline
    return factory

@pytest.fixture
def timed_predicate(timed_predicate_factory):
    return timed_predicate_factory()

@pytest.fixture
def deadline(num_seconds):
    return flux.current_timeline.time() + num_seconds

@pytest.fixture
def num_seconds():
    return 10


@pytest.fixture
def pred1():
    return Pred()


@pytest.fixture
def pred2():
    return Pred()


@pytest.fixture
def forge(request):

    returned = Forge()

    @request.addfinalizer
    def cleanup():  # pylint: disable=unused-variable
        returned.verify()
        returned.restore_all_replacements()

    return returned


@pytest.fixture(autouse=True, scope='session')
def freeze_flux():
    flux.current_timeline.set_time_factor(0)


@pytest.fixture
def state():
    return State()


@pytest.fixture
def pact(state):
    return Pact('test pact').until(state.is_finished)


@pytest.fixture
def pact_duration(state, num_seconds):
    return Pact('test pact', num_seconds).until(state.is_finished)


class State(object):

    def __init__(self):
        super(State, self).__init__()
        self._finished = False
        self.is_finished_call_count = 0

    def is_finished(self):
        self.is_finished_call_count += 1
        return self._finished

    def finish(self):
        self._finished = True



import itertools

from .base import PactBase


class PactGroup(PactBase):

    def __init__(self, pacts=None, lazy=True, timeout_seconds=None):
        if pacts is None:
            pacts = []
        self._pacts = list(pacts)
        self._finished_pacts = []
        self._is_lazy = lazy
        super(PactGroup, self).__init__(timeout_seconds)

    def __iadd__(self, other):
        self.add(other)
        return self

    def __iter__(self):
        return itertools.chain(self._pacts, self._finished_pacts)

    def get_timeout_exception(self, exc_info):
        if self._pacts:
            return self._pacts[0].get_timeout_exception(exc_info)
        return super(PactGroup, self).get_timeout_exception(exc_info)

    def add(self, pact, absorb=False):
        if absorb and isinstance(pact, PactGroup):
            raise NotImplementedError('Absorbing groups is not supported') # pragma: no cover
        self._pacts.append(pact)
        if absorb:
            # pylint: disable=protected-access
            while pact._then:
                # then might throw, so we attempt it first
                self.then(pact._then[0])
                pact._then.pop(0)
            while pact._lastly:
                # lastly might throw, so we attempt it first
                self.lastly(pact._lastly[0])
                pact._lastly.pop(0)

    def _is_finished(self):
        has_finished = True
        indexes_to_remove = []
        for index, pact in enumerate(self._pacts):
            if pact.poll():
                indexes_to_remove.append(index)
            else:
                has_finished = False
                if self._is_lazy:
                    break
        for index in reversed(indexes_to_remove):
            self._finished_pacts.append(self._pacts.pop(index))
        return has_finished

    def __repr__(self):
        return repr(list(self._pacts))



from __future__ import print_function

import doctest
import os

import pytest

PROJECT_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

DOCS_DIRS = [PROJECT_ROOT, os.path.join(PROJECT_ROOT, 'doc')]

@pytest.mark.parametrize("path", [
    os.path.join(path, filename)
    for path in DOCS_DIRS
    for filename in os.listdir(path)
    if filename.endswith((".rst", ".md"))])
def test_doctests(path):
    assert os.path.exists(path)

    context_filename = path + ".ctx.py"
    context = {}
    if os.path.exists(context_filename):
        with open(context_filename) as f:
            exec(f.read(), context)  # pylint: disable=exec-used

    result = doctest.testfile(path, module_relative=False, globs=context)
    assert result.failed == 0



import pytest
import flux
from pact import Pact, PactGroup
from waiting.exceptions import TimeoutExpired

# pylint: disable=redefined-outer-name

def test_group_wait_during(checkpoint, timed_group, num_seconds):
    timed_group.during(checkpoint)
    timed_group.wait()
    assert checkpoint.called_times == num_seconds + 1


def test_group_wait_then(checkpoint, checkpoint1, checkpoint2, checkpoint3, timed_group, timed_pact):
    timed_group.lastly(checkpoint1)
    timed_group.then(checkpoint)
    timed_pact.then(checkpoint2)
    timed_pact.lastly(checkpoint3)
    timed_group.wait()
    assert checkpoint.called
    assert checkpoint1.called
    assert checkpoint2.called
    assert checkpoint3.called
    assert timed_group.is_finished()


@pytest.mark.parametrize('adding_group', [True, False])
def test_group_iadd_pact(adding_group):
    group = PactGroup()
    if adding_group:
        p1 = Pact('c') + Pact('d')
    else:
        p1 = Pact('c')
    group += p1
    assert group._pacts[-1] is p1  # pylint: disable=protected-access


def test_group_with_duration(pact, num_seconds):
    time_before = flux.current_timeline.time()
    group = PactGroup([pact], timeout_seconds=num_seconds)
    with pytest.raises(TimeoutExpired):
        group.wait()
    assert num_seconds == flux.current_timeline.time() - time_before


def test_group_with_duration_wait_timeout_seconds(pact, num_seconds):
    timeout = num_seconds + 1
    time_before = flux.current_timeline.time()
    group = PactGroup([pact], timeout_seconds=num_seconds)
    with pytest.raises(TimeoutExpired):
        group.wait(timeout_seconds=timeout)
    assert timeout == flux.current_timeline.time() - time_before


def test_group_and_pact_with_duration(pact_duration, num_seconds):
    group_duration = num_seconds + 10
    time_before = flux.current_timeline.time()
    group = PactGroup([pact_duration], timeout_seconds=group_duration)
    with pytest.raises(TimeoutExpired):
        group.wait()
    assert group_duration == flux.current_timeline.time() - time_before


def test_group_with_custom_timeout(state):
    class MyTimeoutException(Exception):
        pass
    class MyPact(Pact):
        def get_timeout_exception(self, exc_info):
            return MyTimeoutException()

    pact = MyPact('custom exception pact').until(state.is_finished)
    group = PactGroup([Pact('dummy pact'), pact], timeout_seconds=0)
    with pytest.raises(MyTimeoutException):
        group.wait()


def test_group_without_absorb(pred1, pred2, checkpoint, checkpoint1, checkpoint2, checkpoint3):
    p1 = Pact('a').until(pred1).lastly(checkpoint).then(checkpoint1)
    p2 = Pact('b').until(pred2).then(checkpoint2).lastly(checkpoint3)
    group = p1 + p2
    assert not checkpoint1.called
    pred1.satisfy()
    assert not group.poll()
    assert checkpoint.called
    assert checkpoint1.called
    assert not checkpoint2.called
    assert not checkpoint3.called
    pred2.satisfy()
    assert group.poll() and group.is_finished()
    assert checkpoint2.called
    assert checkpoint3.called


def test_group_with_absorb(pred1, pred2, checkpoint, checkpoint1, checkpoint2, checkpoint3):
    # pylint: disable=protected-access
    p1 = Pact('a').until(pred1).lastly(checkpoint).then(checkpoint1)
    p2 = Pact('b').until(pred2).then(checkpoint2).lastly(checkpoint3)
    group = PactGroup()
    group.add(p1, absorb=True)
    group.add(p2, absorb=True)
    assert not p1._then
    assert not p1._lastly
    assert not p2._then
    assert not p2._lastly
    assert not checkpoint.called
    assert not checkpoint1.called
    pred1.satisfy()
    assert not group.poll() and not group.is_finished()
    assert not checkpoint.called
    assert not checkpoint1.called
    assert not checkpoint2.called
    assert not checkpoint3.called
    pred2.satisfy()
    assert group.poll() and group.is_finished()
    assert checkpoint.called
    assert checkpoint1.called
    assert checkpoint2.called
    assert checkpoint3.called


def test_abosrbing_group_not_implemented():
    group = PactGroup()
    group2 = PactGroup()
    with pytest.raises(NotImplementedError):
        group.add(group2, absorb=True)


def test_iterating_empty_pact_group():
    assert list(PactGroup()) == []


@pytest.mark.parametrize('satisfy_first', [True, False])
def test_iterating_nonempty_pact_group(pred1, pred2, satisfy_first):
    g = PactGroup()
    p1 = Pact('1').until(pred1)
    p2 = Pact('2').until(pred2)
    g.add(p1)
    g.add(p2)

    assert list(g) == [p1, p2]

    if satisfy_first:
        pred1.satisfy()
    else:
        pred2.satisfy()

    g.poll()
    assert sorted(g, key=str) == [p1, p2]



@pytest.fixture
def timed_group(absorb, timed_pact):
    returned = PactGroup()
    returned.add(timed_pact, absorb=absorb)
    return returned


@pytest.fixture
def timed_pact(timed_predicate_factory):
    return Pact('delayed').until(timed_predicate_factory())


@pytest.fixture(params=[True, False])
def absorb(request):
    return request.param



# -*- coding: utf-8 -*-

# Based on logbook.helpers, licensed under BSD. See https://github.com/mitsuhiko/logbook/blob/0.4.2/logbook/compat.py
# for copyright information

# pylint: disable=import-error
# pylint: disable=maybe-no-member
# pylint: disable=no-name-in-module
# pylint: disable=undefined-variable
# pylint: disable=unused-argument
# pylint: disable=unused-import
# pylint: disable=exec-used
# pylint: disable=redefined-builtin
import sys

PY2 = sys.version_info[0] == 2

if PY2:
    import __builtin__ as _builtins
else:
    import builtins as _builtins

if PY2:
    from cStringIO import StringIO
    iteritems = lambda d: d.iteritems() # not dict.iteritems!!! we support ordered dicts as well
    itervalues = lambda d: d.itervalues()
    from itertools import izip as zip
    xrange = _builtins.xrange
else:
    from io import StringIO
    zip = _builtins.zip
    xrange = range
    iteritems = lambda d: iter(d.items()) # not dict.items!!! See above
    itervalues = lambda d: iter(d.values())

_IDENTITY = lambda obj: obj

if PY2:
    integer_types = (int, long)
    string_types = (basestring,)
else:
    integer_types = (int,)
    string_types = (str,)

if PY2:
    #Yucky, but apparently that's the only way to do this
    exec("""
def reraise(tp, value, tb=None):
    raise tp, value, tb
""", locals(), globals())
else:
    def reraise(tp, value, tb=None):
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value



import itertools
import flux

_async_deletes = {}
_id = itertools.count()

def delete_async(path):
    returned = next(_id)
    timeout = 10 if 'huge' not in path else 30
    _async_deletes[returned] = flux.current_timeline.time() + timeout
    return returned

def is_async_delete_finished(operation_id):
    return flux.current_timeline.time() >= _async_deletes[operation_id]



import pact
import flux
import pytest
from waiting.exceptions import TimeoutExpired

# pylint: disable=redefined-outer-name

@pytest.mark.parametrize('reprify', [repr, str])
def test_str_repr(reprify):
    msg = 'Some message here'
    p = pact.Pact(msg)
    assert msg in reprify(p)


def test_pact_is_finished_doesnt_poll(pact, state):
    assert not pact.is_finished()
    assert state.is_finished_call_count == 0


def test_pact_is_finished_not_called_after_finish_poll(pact, state):
    assert state.is_finished_call_count == 0
    assert not pact.poll()
    assert not pact.poll()
    assert state.is_finished_call_count == 2
    state.finish()
    assert pact.poll()
    assert pact.poll()
    assert state.is_finished_call_count == 3


def test_add_pact_callback_after_pact_finished(pact, state, pact_callback, forge, callback):
    assert not pact.is_finished()
    # expect
    callback(1).and_return(True)

    forge.replay()

    pact_callback(callback, 1)

    state.finish()
    assert pact.poll() and pact.is_finished()
    with pytest.raises(RuntimeError):
        pact_callback(callback)

    forge.verify()


def test_add_not_callable_to_pact_fails(pact_callback):
    with pytest.raises(AssertionError):
        pact_callback(True)


def test_during_exception(pact, state, forge, callback):
    callback(1)
    callback(2).and_raise(SampleException())
    callback(3)
    callback(4)
    callback(5)

    forge.replay()

    pact.during(callback, 1)
    pact.during(callback, 2)
    pact.during(callback, 3)
    pact.then(callback, 4)
    pact.lastly(callback, 5)

    state.finish()
    with pytest.raises(SampleException):
        pact.poll()

    forge.verify()
    assert pact.is_finished()


def test_then_exception(pact, state, forge, callback):
    callback(1)
    callback(2).and_raise(SampleException())
    callback(3)
    callback(4)
    callback(5).and_raise(AnotherSampleException())
    callback(6)

    forge.replay()

    pact.lastly(callback, 4)
    pact.lastly(callback, 5)
    pact.then(callback, 1)
    pact.then(callback, 2)
    pact.lastly(callback, 6)
    pact.then(callback, 3)

    state.finish()
    with pytest.raises(SampleException):
        pact.poll()

    forge.verify()
    assert pact.is_finished()


def test_on_timeout(pact, callback, forge):
    callback(1)
    callback(2)
    pact.on_timeout(callback, 1).on_timeout(callback, 2)

    forge.replay()
    with pytest.raises(TimeoutExpired):
        pact.wait(timeout_seconds=3)


def test_wait_no_timeout_seconds_throws_after_duration(pact_duration, num_seconds):
    time_before = flux.current_timeline.time()
    with pytest.raises(TimeoutExpired):
        pact_duration.wait()
    assert num_seconds == flux.current_timeline.time() - time_before


def test_wait_no_timeout_seconds_called_after_duration_passed_immediately_throws(pact_duration, num_seconds):
    time_first = flux.current_timeline.time()
    flux.current_timeline.set_time(time_first + num_seconds + 1)
    time_before = flux.current_timeline.time()
    with pytest.raises(TimeoutExpired):
        pact_duration.wait()
    assert time_before == flux.current_timeline.time()


def test_wait_with_timeout_seconds_and_duration_throws_after_timeout_seconds(pact_duration, num_seconds):
    timeout = num_seconds + 1
    time_before = flux.current_timeline.time()
    with pytest.raises(TimeoutExpired):
        pact_duration.wait(timeout_seconds=timeout)
    assert timeout == flux.current_timeline.time() - time_before


def test_wait_with_timeout_seconds_and_duration_throws_after_timeout_seconds_duration_smaller(pact_duration,
                                                                                              num_seconds):
    timeout = num_seconds - 1
    time_before = flux.current_timeline.time()
    with pytest.raises(TimeoutExpired):
        pact_duration.wait(timeout_seconds=timeout)
    assert timeout == flux.current_timeline.time() - time_before


def test_duration_doesnt_effect_finished_pacts(pact_duration, state):
    state.finish()
    time_before = flux.current_timeline.time()
    pact_duration.wait()
    assert time_before == flux.current_timeline.time()
    assert pact_duration.poll() and pact_duration.is_finished()


class SampleException(Exception):
    pass


class AnotherSampleException(Exception):
    pass


@pytest.fixture(params=['then', 'until', 'during'])
def pact_callback(request, pact):
    name = request.param
    return getattr(pact, name)



class EdgeTriggered(object):

    def __init__(self, callback, args, kwargs):
        super(EdgeTriggered, self).__init__()
        self._callback = callback
        self._args = args
        self._kwargs = kwargs
        self._satisfied = False

    def satisfied(self):
        if not self._satisfied:
            self._satisfied = self._callback(*self._args, **self._kwargs)
        return self._satisfied



from .utils import EdgeTriggered
from .base import PactBase
from .group import PactGroup


class Pact(PactBase):

    def __init__(self, msg, timeout_seconds=None):
        self.msg = msg
        super(Pact, self).__init__(timeout_seconds)
        self._until = []

    def until(self, predicate, *args, **kwargs):
        """Adds a callback criterion for the completion of this pact

        :param predicate: A callable that should return a True-ish value once the end condition is met

        .. note:: When adding multiple predicates via multiple calls to until(), the pact waits on *all* of them to be
          satisfied
        """
        assert callable(predicate)
        self._validate_can_add_callback()
        self._until.append(EdgeTriggered(predicate, args, kwargs))
        return self

    def _is_finished(self):
        return all(predicate.satisfied() for predicate in self._until)

    def group_with(self, other):
        return PactGroup([self, other])

    def __add__(self, other):
        return self.group_with(other)

    def __repr__(self):
        return '<{0.__class__.__name__}: {0.msg}>'.format(self)
